use std::sync::Arc;
use async_trait::async_trait;
use mailparse::ParsedMail;
use matrix_sdk::Client;
use postbus::{Handler, SmtpService, SmtpState, command::Mailbox};
use crate::config::Mapping;

mod matrix;
mod config;


#[tokio::main]
async fn main() {
    let args: Vec<String> = std::env::args().collect();
    if args.len() != 2 {
        panic!("The path to the configuration file must be supplied as (only) command line argument!");
    }
    let mut config = crate::config::load_config(&args[1]);
    let immutable_config = config.clone();
    let clients = matrix::get_clients(&mut config).await;
    crate::config::write_config(&config, &immutable_config, &args[1]);

    let service = SmtpService::create(format!("{}:{}", config.bind_address.unwrap(), config.bind_port.unwrap()).parse().unwrap(), "matrixmailer".into(), 
        Arc::new(ToMatrixConverter {
            allowed_receivers: config.allowed_receivers,
            matrix_clients: clients,
            mappings: config.mappings,
    }));
    service.listen().await;
}

struct ToMatrixConverter {
    // If None, any incoming mail will be accepted. 
    // If Some(x), any mail with a mail address in x in its recipients will be accepted.
    // "Not accepted" means the sender will get a "RCPT failed" reply
    allowed_receivers: Option<Vec<String>>,
    matrix_clients: Vec<Client>,
    mappings: Vec<Mapping>,
}

#[async_trait]
impl Handler for ToMatrixConverter {
    // Checks whether receipient is local (and the mail should thus be handled by this service)
    async fn recipient_local(&self, recipient: &Mailbox) -> bool {
        match &self.allowed_receivers {
            None => true,
            Some(list) => {
                let addr = format!("{}@{}", recipient.local, recipient.domain.0);
                for entry in list {
                    if entry.eq(&addr) {
                        return true;
                    }
                }
                false
            }
        }
    }
    async fn save(&self, state: &SmtpState) -> bool {
        for recipient in &state.recipients {
            match mailparse::parse_mail(state.data.as_bytes()) {
                Ok(mail) => matrix::send(&state.from, &recipient, 
                    &mail_to_plain_message(&mail, &state.from, &recipient), 
                    &mail_to_html_message(&mail, &state.from, &recipient), 
                    &self.matrix_clients, &self.mappings).await,
                Err(e) => {
                    println!("{}", e);
                    matrix::send(&state.from, &recipient, 
                        &format!("Error parsing e-mail. This is a dump of the whole data section:\n{}", &state.data), 
                        &format!("Error parsing e-mail. This is a dump of the whole data section:<br>{}", &state.data), 
                        &self.matrix_clients, &self.mappings).await;
                }
            }
        }
        true
    }
}

fn mail_to_plain_message(mail: &ParsedMail, from: &Option<Mailbox>, to: &Mailbox) -> String {
    format!("From {} to {}\nSubject: {}\n---\n{}", 
        match from {
            Some(fr) => mailbox_to_addr(fr),
            None => "[unknown sender]".to_string(),
        }, 
        mailbox_to_addr(to), 
        get_subject(mail),
        match mail.get_body() {
            Ok(body) => body,
            Err(e) => "WARNING! E-mail content could not be decoded.".to_string(),
        }
    )
}
fn mail_to_html_message(mail: &ParsedMail, from: &Option<Mailbox>, to: &Mailbox) -> String {
    format!("<b>From</b> {} <b>to</b> {}<br><b>Subject:</b> {}<hline>{}", 
        match from {
            Some(fr) => mailbox_to_addr(fr),
            None => "[unknown sender]".to_string(),
        }, 
        mailbox_to_addr(to), 
        get_subject(mail),
        match mail.get_body() {
            Ok(body) => body,
            Err(e) => "<b>WARNING! E-mail content could not be decoded.<b>".to_string(),
        }
    )
}

fn get_subject(mail: &ParsedMail) -> String {
    for header in &mail.headers {
        if header.get_key() == "Subject" {
            return header.get_value();
        }
    }
    return String::from("[no subject found]");
}
pub fn mailbox_to_addr(mb: &Mailbox) -> String {
    format!("{}@{}", mb.local, mb.domain.0)
}